<link
  rel="shortcut icon"
  href="https://raw.githubusercontent.com/f3d-app/f3d/master/resources/logo.ico"
  type="image/x-icon"
/>
<script
  type="application/javascript"
  src="{{ site.baseurl }}/assets/main.js"
></script>
