#include <f3d/camera.h>
#include <f3d/engine.h>
#include <f3d/exception.h>
#include <f3d/export.h>
#include <f3d/image.h>
#include <f3d/interactor.h>
#include <f3d/log.h>
#include <f3d/options.h>
#include <f3d/scene.h>
#include <f3d/types.h>
#include <f3d/window.h>

int main()
{
  f3d::engine::autoloadPlugins();

  f3d::engine eng = f3d::engine::createNone();

  f3d::log::info("F3D engine is loaded");

  return 0;
}
