#include "vtkF3DOrientationMarkerWidget.h"

#include <vtkObjectFactory.h>

vtkStandardNewMacro(vtkF3DOrientationMarkerWidget);

vtkF3DOrientationMarkerWidget::vtkF3DOrientationMarkerWidget()
{
  this->Interactive = 0;
}
